import re
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver

import jsonfield

import logging
logger = logging.getLogger(__name__)


class BasePXEModel(models.Model):
    params = jsonfield.JSONField(null=True, blank=True)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True, null=True, blank=True)

    class Meta:
        abstract = True


class HostGroup(BasePXEModel):
    name = models.CharField("Name", max_length=255, unique=True)
    smart_proxy = models.ForeignKey('pxetool.SmartProxy', null=False)
    source = models.ForeignKey('HostSource', null=True, blank=True, on_delete=models.SET_NULL)
    source_params = jsonfield.JSONField(null=True, blank=True)

    class Meta:
        app_label = "pxetool"

    def __unicode__(self):
        return self.name

    def __getattr__(self, item):
        if isinstance(self.params, dict) and item in self.params:
            return self.params[item]
        raise AttributeError

    @property
    def config_url(self):
        return self.smart_proxy.pxetool_url

    @property
    def pxe_server(self):
        return self.smart_proxy.hostname

    @property
    def pxe_server_tftpboot(self):
        return self.smart_proxy.tftpboot

    @property
    def repos(self):
        return self.params.get('additional_repos', []) if self.params else []

    @property
    def mirror_url(self):
        return self.mirrors


class Host(BasePXEModel):
    hostname = models.CharField("Hostname", max_length=255)
    mac = models.CharField("MAC address", max_length=17, unique=True)
    profile = models.ForeignKey('pxetool.Profile', related_name='hosts', null=True, on_delete=models.SET_NULL)
    installation = models.ForeignKey('pxetool.Installation', related_name='+', null=True)
    group = models.ForeignKey('HostGroup', related_name='hosts', null=False, on_delete=models.PROTECT)
    source = models.ForeignKey('HostSource', null=True, on_delete=models.SET_NULL, blank=True)

    def __getattr__(self, item):
        if self.params and item in self.params:
            return self.params[item]
        return getattr(super(Host, self), item)

    def __unicode__(self):
        return '{} / {}'.format(self.group.name, self.hostname)

    class Meta:
        app_label = "pxetool"
        unique_together = ('hostname', 'group')
        ordering = ('hostname', )

    @classmethod
    def create(cls, args):
        from pxetool.models import form_profile_host
        profile_form = form_profile_host(data=args)
        host = cls(
            hostname=args['hostname'],
            mac=args['mac'],
            group=HostGroup.objects.get(pk=args['group']),
        )
        if profile_form.is_valid():
            host.params = profile_form.cleaned_data
        host.save()
        from pxetool_drivers.models import DriverManager
        driver_opt = {}
        driver_name = args['driver_name']
        driver_opt['driver_name'] = args['driver_name']
        driver_opt['params'] = {}
        for arg, value in args.items():
            arg_regex = re.match(driver_name + '\-(.*)', arg)
            if arg_regex:
                driver_opt['params'][arg_regex.groups()[0]] = value
        driver = DriverManager(host=host, **driver_opt)
        driver.save()
        return host

    @property
    def domain(self):
        return '.'.join(self.hostname.split('.')[1:])

    @property
    def short_hostname(self):
        return self.hostname.split('.')[0]

    @property
    def get_dashed_mac(self):
        return '01-%s' % re.sub(':', '-', self.mac.lower())

    @property
    def status(self):
        return self.installation.status if self.installation else 'unknown'

    @property
    def color(self):
        st = {
            'pending_installation': 'active',
            'booting_to_installer': 'info',
            'installing': 'info',
            'booting_to_os': 'info',
            'configuring': 'warning',
            'done': 'success',
            'error': 'danger',
            'aborted': 'danger',
            'unknown': '',
        }
        return st[self.status if self.status in st else 'unknown']

    @property
    def installing(self):
        if self.installation.status not in ['done', 'error', 'aborted']:
            return self.installation
        return None

    def reboot(self):
        self.driver.power_cycle(host=self)

    def power_status(self):
        return self.driver.power_state()


class HostSource(BasePXEModel):
    name = models.CharField(max_length=255)
    url = models.CharField(max_length=255)
    type = models.CharField(
        max_length=10,
        choices=map(lambda x: (x, x), ('libvirt', 'inventory')),
        default='inventory'
    )
    enable = models.BooleanField(default=True)

    class Meta:
        app_label = "pxetool"
        ordering = ('name', )

    def __unicode__(self):
        return "{} (URL: {})".format(self.name, self.url)

    @property
    def x_auth_application(self):
        return self.params.get('X-Auth-Application', '') if self.params else ''

    @property
    def x_auth_token(self):
        return self.params.get('X-Auth-Token', '') if self.params else ''


@receiver(post_save)
def post_save_host_source(sender, instance, **kwargs):
    if sender not in [HostSource, HostGroup]:
        return
    from pxetool.tasks import fetch_hosts_from_source
    fetch_hosts_from_source()
