import os
import datetime
import uuid

from django.db import models
from django.template import Template, Context
from django.db.models.signals import pre_delete, pre_save, post_save
from django.dispatch import receiver
from django.core.urlresolvers import reverse

from jsonfield import JSONField
from django.contrib.auth.models import User

import logging
from pxetool.models.puppet import PuppetMaster
from pxetool.exceptions import InstallationException

logger = logging.getLogger(__name__)
status_choices = list(map(lambda x: (x, x), (
    'pending_installation',
    'booting_to_installer',
    'installing',
    'booting_to_os',
    'configuring',
    'sign_request',
    'done',
    'error',
    'aborted',
    'unknown',
)))


class BaseModel(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True
        ordering = ('-pk', )
        app_label = "pxetool"


class InstallationContext(Context):
    def __init__(self, *args, **kwargs):
        self.installation = kwargs.pop('installation')
        super(InstallationContext, self).__init__(*args, **kwargs)
        self.dicts.append({'install': self.installation})

    def __getitem__(self, item):
        # looking up for variable in:
        # - installation
        # - host
        # - group
        # - profile defaults
        try:
            return super(InstallationContext, self).__getitem__(item)
        except KeyError:
            sources = [
                self.installation,
                self.installation.host,
                self.installation.host.group,
                self.installation.profile,
            ]
            for i in sources:
                if hasattr(i, item):
                    return getattr(i, item)
            raise


class Installation(BaseModel):
    owner = models.ForeignKey(User, null=True, on_delete=models.SET_NULL)
    uuid = models.CharField(max_length=36, unique=True)
    profile = models.ForeignKey('pxetool.Profile', related_name='installation',
                                null=True, on_delete=models.SET_NULL)
    profile_name = models.CharField(max_length=255)
    host = models.ForeignKey('pxetool.Host', related_name='installations')
    status = models.CharField(choices=status_choices, max_length=20, default='unknown')
    params = JSONField(null=True)

    # Local copy of smart-proxy in case of changes
    smart_proxy = models.ForeignKey("pxetool.SmartProxy", null=True, on_delete=models.SET_NULL)

    # Host's IP address at the time of installation.
    # Used as a minor security measure in pxetool_ui.views.CfgView
    ip = models.CharField('Host IP Address', null=True, max_length=45)

    # kept for compatibility with existing profiles
    @property
    def hash(self):
        return self.uuid

    @property
    def install_url(self):
        """
        Used as base url for getting files for installations.
        Example: {{install_url}}/firstboot.sh
        """
        return "{pxetool_url}{base_url}{uuid}".format(
            pxetool_url=self.smart_proxy.pxetool_url.rstrip('/'),
            base_url=reverse('install'),
            uuid=self.uuid,
        )

    @property
    def install_status_url(self):
        return "{pxetool_url}{status_url}".format(
            pxetool_url=self.smart_proxy.pxetool_url.rstrip('/'),
            status_url=reverse('install_status', kwargs={'hash': self.uuid}),
        )

    def delete_pxe(self):
        from pxetool.tasks import delete_file_task
        # installation may be tied to the deleted smart-proxy
        # that is why we have this condition
        if self.smart_proxy:
            pxelinux_cfg = os.path.join(self.smart_proxy.tftpboot, 'pxelinux.cfg',
                                        self.host.get_dashed_mac)
            delete_file_task(self.smart_proxy_id, pxelinux_cfg)
        return True

    def install(self):
        from pxetool.tasks import \
            upload_profile_task, \
            upload_file_content_task, \
            power_cycle_task

        host = self.host
        pxelinux_cfg = os.path.join(
            self.smart_proxy.tftpboot,
            'pxelinux.cfg',
            host.get_dashed_mac)
        pxelinux_content = self.files.get(filename='pxelinux.cfg').content

        task = (
            upload_profile_task.si(self.smart_proxy_id, self.profile_id, install_pk=self.pk) |
            upload_file_content_task.si(self.smart_proxy_id, pxelinux_content, pxelinux_cfg, overwrite=True,
                                        install_pk=self.pk) |
            power_cycle_task.si(host.pk, install_pk=self.pk)
        )
        task()
        self.update_status('pending_installation')
        self.save()

    def abort(self):
        if self.delete_pxe():
            self.update_status('aborted')
            self.save()

    def update_status(self, status):
        status = status if status in [i[0] for i in status_choices] else 'unknown'
        # ToDo: remove hardcoded 'puppet_server' variable
        for action in self.actions.filter(status=status):
            logger.debug("Execute action: {}".format(action.action))
            if action.action == 'puppet_cert_clean':
                pm = PuppetMaster.objects.filter(
                    hostname=self.host.group.puppet_server).first()
                if pm:
                    pm.clean_cert(self.host.hostname)
            elif action.action == 'puppet_cert_sign':
                pm = PuppetMaster.objects.filter(
                    hostname=self.host.group.puppet_server).first()
                if pm:
                    pm.sign_cert(self.host.hostname)
            elif action.action == 'delete_pxe':
                self.delete_pxe()
        logs = InstallLog.objects.filter(installation=self)
        if logs:
            last_status = logs.last().status_to
        else:
            last_status = 'unknown'
        log = InstallLog(installation=self,
                         status_from=last_status,
                         status_to=status)
        log.save()
        self.status = status
        self.save()

    def times(self):
        # make more optimal elements placement
        logs = InstallLog.objects.filter(installation=self)
        log_list = []
        log = None
        if logs.count() == 0:
            return log_list
        created = logs[0].created
        for log in logs:
            log_list.append({
                'timer': "{:5d}s".format((log.created - created).seconds),
                'message': '{} => {}'.format(log.status_from, log.status_to),
            })
        if logs[logs.count() - 1].status_to not in {'done', 'error', 'aborted', 'unknown'}:
            log_list.append({
                'timer': "{:5d}s".format((datetime.datetime.now() - created.replace(tzinfo=None)).seconds),
                'message': '{} => {}'.format(log.status_to, ''),
            })
        return log_list


class InstallationFile(models.Model):
    install = models.ForeignKey('Installation', related_name='files')
    filename = models.CharField(max_length=255, db_index=True)
    status = models.CharField(choices=status_choices, max_length=20, null=True)
    content = models.TextField()

    class Meta:
        app_label = "pxetool"


class InstallLog(BaseModel):
    installation = models.ForeignKey('Installation', related_name='logs')
    status_from = models.CharField(choices=status_choices, max_length=20, default='unknown')
    status_to = models.CharField(choices=status_choices, max_length=20, default='unknown')
    action = models.CharField(max_length=255, null=False)

    class Meta:
        ordering = ('created', )
        app_label = "pxetool"

    @classmethod
    def create(cls, installation, action):
        install = cls(installation=installation)
        install.action = action
        install.save()


class LogFile(BaseModel):
    installation = models.ForeignKey('Installation', related_name='log_file')
    filename = models.CharField(max_length=255, null=False)
    content = models.TextField(blank=True, null=False)


class InstallAction(models.Model):
    install = models.ForeignKey('Installation', related_name='actions')
    status = models.CharField(choices=status_choices, max_length=20, default='unknown')
    action = models.CharField(max_length=50, null=False)

    class Meta:
        unique_together = ("install", "action")
        app_label = "pxetool"


@receiver(pre_delete, sender=Installation)
def cleanup(sender, instance, **kwargs):
    """Remove PXE, preseed files before deletings host"""
    if instance.status not in {'done', 'abort'}:
        instance.delete_pxe()


@receiver(pre_save, sender=Installation)
def pre_create(sender, instance, **kwargs):
    # check for a new installation
    if not instance.uuid:
        if not instance.profile:
            # if profile is None, then we can't do anything
            raise InstallationException("Can't start installation without profile")
        # generate random uuid for installation
        instance.uuid = uuid.uuid4()
        while Installation.objects.filter(uuid=instance.uuid):
            instance.uuid = uuid.uuid4()


@receiver(post_save, sender=Installation)
def post_create(sender, instance, **kwargs):
    # we should not do anything on update
    if not kwargs['created']:
        return

    # make a local copy of variables
    instance.profile_name = instance.profile.name
    instance.smart_proxy = instance.host.group.smart_proxy
    instance.params = instance.host.params

    # update reference to last installation on host model
    instance.host.installation = instance
    instance.host.save()

    # we should copy actions in case profile
    # changed during installation
    for action in instance.profile.actions.all():
        install_action = InstallAction(
            action=action.action,
            status=action.status,
            install=instance,
        )
        install_action.save()

    # pregenerate all templates for installation
    for profile_file in instance.profile.files.all():
        content = profile_file.content
        template = Template(content)
        InstallationFile.objects.create(
            install=instance,
            filename=profile_file.filename,
            status=profile_file.status,
            content=template.render(
                InstallationContext(installation=instance)
            )
        )
    instance.install()
