import os

from django.db import models
from django import forms
from django.core.exceptions import ObjectDoesNotExist
from collections import OrderedDict
import tarfile
import tempfile
import yaml
import shutil
import jsonfield

from pxetool.exceptions import ImportProfileException
from django.conf import settings

PROFILES_FILES = settings.PROFILES_FILES


class ProfileField(models.Model):
    field_name = models.CharField(max_length=100)
    field = models.CharField(max_length=50, default='CharField')
    widget = models.CharField(max_length=50, null=True)
    params = jsonfield.JSONField()
    visible_in = models.CharField(max_length=12, default='None', choices=(
        ('None', 'None'), ('host_fields', 'host_fields'), ('group_fields', 'group_fields')
    ))

    class Meta:
        app_label = "pxetool"

    def __unicode__(self):
        return self.field_name


# ToDo: remove
def __form_profile(group, **kwargs):
    form = forms.Form(**kwargs)
    form.fields = OrderedDict()
    fields = ProfileField.objects.filter(visible_in=group)
    for field in fields:
        field_obj = getattr(forms, field.field)
        kwargs = {}
        if field.widget:
            kwargs['widget'] = getattr(forms, field.widget)
        if field.params:
            for key, value in field.params.items():
                kwargs[key] = value
        kwargs['required'] = False
        form.fields[field.field_name] = field_obj(**kwargs)
    return form


# ToDo: remove
def form_profile_host(**kwargs):
    return __form_profile('host_fields', **kwargs)


# ToDo: remove
def form_profile_group(**kwargs):
    return __form_profile('group_fields', **kwargs)


class Profile(models.Model):
    name = models.CharField(max_length=255, unique=True)
    profile_fields = models.ManyToManyField(ProfileField, related_name='profiles')
    kernel = models.CharField(max_length=255, null=True)
    initrd = models.CharField(max_length=255, null=True)

    class Meta:
        app_label = "pxetool"

    def __unicode__(self):
        return self.name

    def __getattr__(self, item):
        try:
            pf = ProfileField.objects.get(field_name=item)
            if pf:
                if pf.params and 'initial' in pf.params:
                    return pf.params['initial']
                return None
        except ObjectDoesNotExist:
            raise AttributeError('No such key {}'.format(item))

    def metadata_validate(self, filename):
        try:
            with open(filename) as f:
                config = yaml.load(f)
        except IOError as e:
            raise ImportProfileException("Can't read metadata.yaml: {}".format(e))
        except yaml.YAMLError as e:
            raise ImportProfileException("Error in configuration file: {}".format(e))
        if not config:
            raise ImportProfileException("Wrong configuration '{}' file in profile".format(
                os.path.basename(filename)
            ))

        profile_tmp_dir = os.path.dirname(filename)
        profile_dir = os.path.join(PROFILES_FILES, self.name)
        # validate destination dir
        if os.path.exists(profile_dir):
            if not os.path.isdir(profile_dir):
                raise ImportProfileException(
                    "Can't create profile dir: {} already exists and not a directory".format(profile_dir)
                )

        # Validate that every template exists in profile
        if 'templates' in config:
            if type(config['templates']) != dict:
                raise ImportProfileException('Wrong template configuration in {}'.format(
                    os.path.basename(filename)
                ))
            for name, template in config['templates'].iteritems():
                if 'source' not in template:
                    raise ImportProfileException("No source for template {}".format())
                template_src = os.path.join(profile_tmp_dir, template['source'])
                if not os.path.exists(template_src):
                    raise ImportProfileException("Can't find template '{}''".format(
                        template['source']
                    ))

        for f in ['kernel', 'initrd']:
            if f not in config:
                continue
            f_src = os.path.join(profile_tmp_dir, config[f])
            if os.path.exists(f_src):
                continue
            raise ImportProfileException("Can't find {} or {} in profile".format(
                os.path.basename(f_src)
            ))

    def import_metadata(self, filename):
        if not self.pk:
            self.save()
        try:
            with open(filename) as f:
                config = yaml.load(f)
        except IOError as e:
            raise ImportProfileException("Can't read metadata.yaml: {}".format(e))
        except yaml.YAMLError as e:
            raise ImportProfileException("Error in configuration file: {}".format(e))
        if not config:
            raise ImportProfileException("Wrong configuration '{}' file in profile".format(
                os.path.basename(filename)
            ))

        profile_tmp_dir = os.path.dirname(filename)
        profile_dir = os.path.join(PROFILES_FILES, self.name)
        # validate destination dir
        if not os.path.exists(profile_dir):
            try:
                os.mkdir(profile_dir, 0o755)
            except (IOError, OSError) as e:
                raise ImportProfileException("Unable to create {}: {}".format(profile_dir, e))

        for f in ['kernel', 'initrd']:
            if f not in config:
                continue
            f_src = os.path.join(profile_tmp_dir, config[f])
            f_dst = os.path.join(profile_dir, config[f])
            try:
                if os.path.exists(f_dst):
                    os.unlink(f_dst)
                shutil.copy(f_src, f_dst)
                os.chmod(f_dst, 0o644)
            except (IOError, OSError) as e:
                raise ImportProfileException("Unable to save '{}' to '{}': {}".format(
                    os.path.basename(f_src), profile_dir, e
                ))
            if hasattr(self, f):
                setattr(self, f, os.path.join(self.name, os.path.basename(f_dst)))

        for group_name in ['group_fields', 'host_fields']:
            if group_name not in config:
                continue
            for i in config[group_name]:
                field_name, params = i.keys()[0], i.values()[0]
                profile_field, created = ProfileField.objects.get_or_create(
                    field_name=field_name,
                )
                for key, value in params.items():
                    setattr(profile_field, key, value)
                profile_field.visible_in = group_name
                profile_field.save()
                self.profile_fields.add(profile_field)

        if 'templates' in config:
            for filename, file_obj in config['templates'].items():
                source = file_obj['source']
                status = file_obj['status'] if 'status' in file_obj else None
                cfg_file, created = ConfigFile.objects.get_or_create(
                    profile=self,
                    filename=filename,
                )
                with open(os.path.join(profile_tmp_dir, source)) as f:
                    cfg_file.content = f.read()
                cfg_file.status = status
                cfg_file.save()
        if 'actions' in config:
            for status, action in config['actions'].items():
                self.actions.get_or_create(status=status, action=action[0])
        self.save()

    def import_archive(self, filename, validate=False):
        temp_dir = tempfile.mkdtemp()
        try:
            tar = tarfile.open(filename, "r:gz")
        except (IOError, OSError, tarfile.TarError) as e:
            raise ImportProfileException('Error opening tar file: {}'.format(e.message))
        try:
            tar.extractall(path=temp_dir)
        except tarfile.TarError as e:
            tar.close()
            raise ImportProfileException('Error extracting tar file: {}'.format(e.message))
        tar.close()

        metadata_path = os.path.join(temp_dir, 'metadata.yaml')
        self.metadata_validate(metadata_path)
        if not validate:
            self.import_metadata(metadata_path)


class ConfigFile(models.Model):
    filename = models.CharField(max_length=255, db_index=True)
    content = models.TextField(null=True)
    profile = models.ForeignKey('Profile', related_name='files')
    status = models.CharField(max_length=20, null=True)

    class Meta:
        app_label = "pxetool"

    def __unicode__(self):
        return self.filename


class ProfileAction(models.Model):
    profile = models.ForeignKey('Profile', related_name='actions')
    status = models.CharField(max_length=20, default='unknown')
    action = models.CharField(max_length=50)

    def __unicode__(self):
        return self.status

    class Meta:
        unique_together = ("profile", "action")
        app_label = "pxetool"
