import os
import json
import requests

from requests.exceptions import ConnectionError, SSLError
from django.db import models
from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver
from django.conf import settings

from logging import getLogger
logger = getLogger(__name__)


class PuppetMasterException(Exception):
    pass


class PuppetMaster(models.Model):
    hostname = models.CharField(max_length=255, db_index=True,
                                default='localhost')
    port = models.IntegerField(default=8140)
    version = models.CharField(max_length=2, choices=(
        ('v1', 'v1'),
    ), default='v1')
    ca_content = models.TextField()
    cert_content = models.TextField()
    key_content = models.TextField()
    status = models.CharField(
        max_length=5, default='Ok',
        choices=(('Ok', 'Ok'), ('Error', 'Error'))
    )

    class Meta:
        app_label = "pxetool"

    @property
    def ca_filename(self):
        return os.path.join(settings.PXETOOL_BASEDIR, "{}_ca.crt"
                            .format(self.hostname))

    @property
    def cert_filename(self):
        return os.path.join(settings.PXETOOL_BASEDIR, "{}.crt"
                            .format(self.hostname))

    @property
    def key_filename(self):
        return os.path.join(settings.PXETOOL_BASEDIR, "{}.key"
                            .format(self.hostname))

    def clean_cert(self, cert_name):
        url = "https://{}:{}/production/certificate_status/{}" \
            .format(self.hostname, self.port, cert_name)
        resp = requests.put(
            url,
            cert=(self.cert_filename, self.key_filename),
            headers={'Content-Type': 'text/pson'},
            data=json.dumps({"desired_state": "revoked"}),
            verify=self.ca_filename,
        )
        # delete only if certificate is revoked
        if resp.content == 'null':
            resp = requests.delete(
                url,
                cert=(self.cert_filename, self.key_filename),
                verify=self.ca_filename,
            )
            if 'Deleted for {}: Puppet::SSL::Certificate' \
                    .format(cert_name) == resp.content:
                return True
        return False

    def sign_cert(self, cert_name):
        url = "https://{}:{}/production/certificate_status/{}" \
            .format(self.hostname, self.port, cert_name)
        return requests.put(
            url,
            cert=(self.cert_filename, self.key_filename),
            headers={'Content-Type': 'text/pson'},
            data=json.dumps({"desired_state": "signed"}),
            verify=self.ca_filename,
        )

    def list_certificates(self):
        url = "https://{}:{}/production/certificate_statuses/:any_key" \
            .format(self.hostname, self.port)
        try:
            data = requests.get(
                url,
                cert=(self.cert_filename, self.key_filename),
                headers={'Accept': 'pson'},
                verify=self.ca_filename,
            )
        except (ConnectionError, SSLError) as e:
            raise PuppetMasterException(
                'Connection error to host "{}":  {}'.format(self.hostname, e)
            )
        try:
            return data.json()
        except ValueError as e:
            raise PuppetMasterException(
                'Unable to list certificates on "{}":  {}'
                .format(self.hostname, e)
            )


@receiver(post_save, sender=PuppetMaster)
def update_certs(sender, instance, **kwargs):
    instance.status = 'Ok'
    try:
        with open(instance.ca_filename, 'w') as f:
            f.write(instance.ca_content)
        with open(instance.cert_filename, 'w') as f:
            f.write(instance.cert_content)
        with open(instance.key_filename, 'w') as f:
            f.write(instance.key_content)
        try:
            instance.list_certificates()
        except PuppetMasterException as e:
            logger.error('PuppetMaster: {}'.format(e.message))
            instance.status = 'Error'
    except IOError as e:
        logger.error('PuppetMaster: Unable to write file to {}: {}'.format(e.filename, e))
        instance.status = 'Error'
    post_save.disconnect(update_certs, sender=PuppetMaster)
    instance.save()
    post_save.connect(update_certs, sender=PuppetMaster)


@receiver(post_delete, sender=PuppetMaster)
def delete_certs(sender, instance, **kwargs):
    try:
        os.remove(instance.ca_filename)
        os.remove(instance.cert_filename)
        os.remove(instance.key_filename)
    except IOError as e:
        logger.error('Unable to delete file {}'.format(e.filename))
