import os
import pwd

from django.test import TestCase
from django.db import IntegrityError
from django.contrib.auth.models import User
from .models import Host, HostGroup, SmartProxy, Profile
from .models import Installation


class HostsTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create_superuser(
            email='user@host.local',
            username='user',
            password='user',
        )
        self.profile = Profile.objects.create(name='test-profile')
        self.profile.files.create(filename='pxelinux.cfg')
        self.sm = SmartProxy.objects.create(
            hostname='localhost',
            username=pwd.getpwuid(os.getuid()).pw_name,
        )
        self.host_groups = [
            HostGroup.objects.create(name='test_group{}'.format(i), smart_proxy=self.sm)
            for i in range(1, 3)
        ]
        self.hosts = [
            Host.objects.create(
                hostname='hg1-host{}.test.local'.format(i),
                mac='01-01-01-01-01-{:02d}'.format(i),
                group=self.host_groups[0],
            )
            for i in range(1, 3)
        ]
        self.hosts.extend([
            Host.objects.create(
                hostname='hg2-host{}.test.local'.format(i),
                mac='01-01-01-01-02-{:02d}'.format(i),
                group=self.host_groups[1],
            )
            for i in range(1, 3)
        ])

    def test_class_to_string(self):
        """Test correct __unicode__ function"""
        self.assertEqual(str(self.host_groups[0]), self.host_groups[0].name)
        self.assertEqual(
            str(self.hosts[0]),
            '{} / {}'.format(self.hosts[0].group.name, self.hosts[0].hostname)
        )

    def test_same_host_different_group(self):
        src_host = Host.objects.filter(group=self.host_groups[0]).first()
        Host.objects.create(
            hostname=src_host.hostname,
            mac='00-01-02-03-04-06',
            group=self.host_groups[1]
        )

    def test_same_host_same_group(self):
        with self.assertRaises(IntegrityError):
            Host.objects.create(hostname='hg1-host1.test.local', mac='00-01-02-03-04-07', group=self.host_groups[0])

    def test_color_property(self):
        st = {
            'pending_installation': 'active',
            'booting_to_installer': 'info',
            'installing': 'info',
            'booting_to_os': 'info',
            'configuring': 'warning',
            'done': 'success',
            'error': 'danger',
            'aborted': 'danger',
            'unknown': '',
            '__nonexistent_item___': '',
        }
        host = Host()
        self.assertEqual(host.status, 'unknown')
        host = self.hosts[0]
        installation = Installation(host=host, owner=self.user, profile=self.profile)
        installation.save()
        host.installation = installation
        host.save()
        for k, v in st.iteritems():
            installation.status = k
            installation.save()
            self.assertEqual(host.color, v)

    def assertRaisesWithArgs(self, excClass, excArgs, callableObj, *args, **kwargs):
        """
        Asserts that callable object will raise specific exception class with certain arguments
        when called with provided `args` and `kwargs`.

        :param excClass: expected exception class to be raised
        :param excArgs: arguments that raised exception is expected to have
        :param callableObj: callable to call
        :param args: positional arguments for callable
        :param kwargs: named arguments for callable
        :return: raises AssertionError if callable doesn't behave as expected
        """
        try:
            callableObj(*args, **kwargs)
        except excClass as e:
            self.assertEqual(excArgs, e.args)
        except Exception as e:
            self.fail('callable {2!r} raised exception type {1}, expected {0}'.format(
                excClass.__name__, type(e).__name__, callableObj
            ))
        else:
            self.fail('callable {2!r} did not raise exception, expected {0} with args {1!r}'.format(
                excClass.__name__, excArgs, callableObj
            ))

    def test_detect_installation_error_decorator(self):
        # Preparation
        from tasks import detect_installation_error

        expected_exception = ZeroDivisionError
        expected_args = (1, 2, 3)

        @detect_installation_error
        def mock_task_fail(*args, **kwargs):
            raise expected_exception(*expected_args)

        @detect_installation_error
        def mock_task_succeed(*args, **kwargs):
            pass

        installation = Installation(host=self.hosts[0], owner=self.user, profile=self.profile)
        installation.status = 'pending_installation'
        installation.save()
        pk = installation.pk

        # Testing

        # Test 1. Check @detect_installation_error behavior for successful tasks.
        # Expected result: no exceptions, no change to installation status.
        mock_task_succeed(install_pk=pk)
        installation = Installation.objects.get(pk=pk)
        self.assertEqual('pending_installation', installation.status)

        # Test 2. Check @detect_installation_error behavior for failed tasks with no pk provided.
        # Expected result: exception propagated transparently, decorator doesn't fail.
        self.assertRaisesWithArgs(expected_exception, expected_args, mock_task_fail)
        installation = Installation.objects.get(pk=pk)
        self.assertEqual('pending_installation', installation.status)

        # Test 3. Check @detect_installation_error behavior for failed tasks.
        # Expected result: exception propagated transparently, status changed to 'error'.
        self.assertRaisesWithArgs(expected_exception, expected_args, mock_task_fail, install_pk=pk)
        installation = Installation.objects.get(pk=pk)
        self.assertEqual('error', installation.status)

        # Test 4. Check @detect_installation_error behavior for failed tasks with deleted install.
        # Expected result: exception propagated transparently, decorator doesn't fail.
        installation.delete()
        self.assertRaises(Installation.DoesNotExist, Installation.objects.get, pk=pk)
        self.assertRaisesWithArgs(expected_exception, expected_args, mock_task_fail, install_pk=pk)
