import importlib
import logging

from django.db import models
from django.conf import settings
from django import forms
import jsonfield

from pxetool.models import Host

logger = logging.getLogger(__name__)
driver_list = settings.DRIVER_LIST

LIB_NAME = '.'.join(__name__.split('.')[:-1])


class DriverManager(models.Model):
    """Driver proxy model
       driver_name : None or driver_name
       driver      : real driver object
       params      : {hash} of driver params
    """
    driver = None
    driver_name = models.CharField(max_length=100, null=True)
    params = jsonfield.JSONField()
    host = models.OneToOneField(Host, related_name='driver')

    def __unicode__(self):
        return self.driver_name if self.driver_name else 'Not specified'

    def __getattribute__(self, item):
        if item == 'get_form':
            return object.__getattribute__(self, '_get_form')
        return object.__getattribute__(self, item)

    def __getattr__(self, item):
        allowed_functions = [
            'power_cycle',
            'power_on',
            'power_off',
            'power_state',
            'screenshot',
            'show_links',
        ]
        if item in allowed_functions and self.driver_name in driver_list:
            if not self.driver:
                self._init_driver()
            return getattr(self.driver, item)
        elif self.driver_name in ['powerless', ''] and item in allowed_functions:
            return lambda: ''
        else:
            raise AttributeError(
                'DriverManager: attribute "{}" not found for driver "{}"'
                .format(item, self.driver_name)
            )

    def _init_driver(self):
        try:
            mod = importlib.import_module('{}.{}'.format(LIB_NAME, self.driver_name))
            driver_class = getattr(mod, self.driver_name.capitalize())
            self.driver = driver_class(self.host, **dict(self.params))
        except ImportError as e:
            logger.error("Failed to import {}".format(e.message))

    @property
    def prefix_params(self):
        """Used for pass params to form"""
        params = {}
        for key, value in self.params.items():
            params[self.driver_name + '-' + key] = value
        return params

    def _get_form(self, *args, **kwargs):
        return DriverManager.get_form(*args, self=self, **kwargs)

    @staticmethod
    def get_form(*args, **kwargs):
        """Used in host.views for checking forms"""

        if 'self' in kwargs:
            self = kwargs.pop('self')
            kwargs['data'] = self.prefix_params
            driver_name = self.driver_name
        else:
            driver_name = kwargs.pop('driver_name') if 'driver_name' in kwargs else args[0]
            if 'data' not in kwargs:
                kwargs['data'] = dict()
        if not kwargs['data']:
            kwargs.pop('data')
        if driver_name in driver_list:
            try:
                mod = importlib.import_module('{}.{}'.format(LIB_NAME, driver_name))
                driver_class = getattr(mod, driver_name.capitalize())
                form = driver_class.get_form(*args, **kwargs)
            except ImportError as e:
                logger.error("Failed to import {}".format(e.message))
                # Return empty non-valid form.
                form = forms.Form()
        else:
            # Return empty valid form to ensure "powerless" hosts can be successfully submitted.
            form = forms.Form(prefix='powerless', data={})
        return form
