from django.db import models
from django import forms

import libvirt
from datetime import datetime

from pxetool.exceptions import DriverException
from .base_driver import BaseDriver


class VirshForm(forms.Form):
    conn_string = forms.CharField(
        max_length=512,
        help_text='<br />qemu:///system for localhost<br />'
                  'qemu+ssh://root@server.name/system for remote'
    )


class VirshCache(models.Model):
    """Cached info on vm power states"""
    conn_string = models.CharField(max_length=255, default='', null=True)
    last_updated = models.DateTimeField(default=datetime.now)


class Virsh(BaseDriver):
    """Libvirt driver"""

    def __init__(self, *args, **kwargs):
        self.domain = None
        self.conn = None
        (self.host,) = args
        if 'conn_string' not in kwargs:
            raise DriverException("No conn_string parameter")
        self.conn_string = kwargs['conn_string']
        if self.host:
            self.vm_name = self.host.hostname

    def init_conn(self):
        try:
            self.conn = libvirt.open(self.conn_string)
        except libvirt.libvirtError:
            raise DriverException("Unable to connect to libvirt daemon: {}".format(self.conn_string))
        try:
            self.domain = self.conn.lookupByName(self.vm_name) if self.host else None
        except libvirt.libvirtError as e:
            if e.get_error_code() == libvirt.VIR_ERR_NO_DOMAIN:
                raise DriverException("Unable to find Libvirt Domain: {}".format(self.vm_name))
            else:
                pass

    @property
    def screenshot(self):
        if not self.host:
            return
        if not self.conn:
            self.init_conn()
        if self.domain.state()[0] != libvirt.VIR_DOMAIN_RUNNING:
            return False
        from PIL import Image
        import io

        def handler(stream, buf, opaque):
            fd = opaque
            return fd.write(buf)
        stream = self.conn.newStream()
        self.domain.screenshot(stream, 0)
        iostream = io.BytesIO()
        stream.recvAll(handler, iostream)

        img = Image.open(iostream)
        return img

    def power_on(self, *args, **kwargs):
        if not self.host:
            return
        if not self.conn:
            self.init_conn()
        if not self.domain:
            return None
        state = self.domain.state()[0]
        if state == libvirt.VIR_DOMAIN_RUNNING and 'cycle' in kwargs and kwargs['cycle']:
            self.domain.destroy()
        self.domain.create()

    def power_cycle(self):
        if not self.host:
            return
        if not self.conn:
            self.init_conn()
        if not self.domain:
            return None
        state = self.domain.state()[0]
        if state == libvirt.VIR_DOMAIN_PAUSED:
            self.domain.destroy()
            self.domain.create()
        elif state in [libvirt.VIR_DOMAIN_SHUTOFF, libvirt.VIR_DOMAIN_SHUTDOWN]:
            self.domain.create()
        self.domain.reset()

    @property
    def show_links(self):
        return ''

    @property
    def power_state(self):
        """ Returns power state
        :return: [0-off, 1-on, 2-unknown, 3-paused]
        """
        if not self.host:
            return
        if not self.conn:
            self.init_conn()
        if not self.domain:
            return None
        states = {
            libvirt.VIR_DOMAIN_SHUTOFF: 0,
            libvirt.VIR_DOMAIN_SHUTDOWN: 0,
            libvirt.VIR_DOMAIN_RUNNING: 1,
            libvirt.VIR_DOMAIN_PAUSED: 3,
        }
        domain_state = self.domain.state()[0]
        state = 2  # defaults to unknown
        if domain_state in states:
            state = states[domain_state]
        return state

    @property
    def power_color(self):
        if not self.host:
            return
        if not self.conn:
            self.init_conn()
        if not self.domain:
            return None
        state = self.domain.state()[0]
        if state == libvirt.VIR_DOMAIN_SHUTOFF:
            return "red"
        elif state == libvirt.VIR_DOMAIN_RUNNING:
            return "green"
        return "gray"

    def list_domains(self):
        import xml.etree.ElementTree as ET
        domains = {}
        if not self.conn:
            self.init_conn()
        for domain in self.conn.listAllDomains():
            et = ET.fromstring(domain.XMLDesc())
            name = et.find('name').text
            try:
                domains[name] = {'interfaces': []}
                for child in et.find('devices').getchildren():
                    if child.tag == 'interface':
                        domains[name]['interfaces'].append(
                            child.find('mac').attrib['address']
                        )
            except (AttributeError, KeyError):
                # wrong xml config, parse next
                if name in domains[name]:
                    del domains[name]
        return domains

    @staticmethod
    def get_form(prefix='virsh', *args, **kwargs):
        return VirshForm(prefix=prefix, **kwargs)
