import re
import os
import pwd

from django import forms
from django.conf import settings

from django.core.exceptions import ValidationError
from pxetool.models import Host, HostGroup, HostSource, PuppetMaster

from pxetool.models import Profile, ProfileField
from pxetool.models import SmartProxy


class MacField(forms.CharField):
    regex = re.compile(
        "(([0-9a-f]{2}[-:\.]){5}[0-9a-f]{2}|([0-9a-f]{4}\.){2}[0-9a-f]{4})",
        re.IGNORECASE,
    )

    def to_python(self, value):
        if not value:  # if no MAC field submitted
            return ''
        if self.regex.match(value):
            mac_split = re.findall('[0-9a-f]{2}', value, re.I)
            return '-'.join(mac_split)
        return value

    def validate(self, value):
        """Check mac format XX-XX-XX-..."""
        super(MacField, self).validate(value)
        if not self.regex.match(value):
            raise ValidationError(
                "MAC address %s is not valid" % value,
                code='invalid',
                params={'mac': value},
            )


class PuppetMasterField(forms.TypedChoiceField):

    def __init__(self, *args, **kwargs):
        super(PuppetMasterField, self).__init__(*args, **kwargs)
        self.choices = [('', '-' * 10), ]
        self.choices.extend([
            (i.hostname, i.hostname)
            for i in PuppetMaster.objects.all()
        ])
        self.required = False


def unique_checker_mac(value):
    if Host.objects.filter(mac=value):
        raise ValidationError("MAC already used")


def validate_mirror_url(value):
    if not settings.MIRROR_REGEX.match(value):
        raise ValidationError("Wrong mirror url")


class HostForm(forms.Form):
    group = forms.ModelChoiceField(queryset=HostGroup.objects.all(), empty_label=None)
    hostname = forms.CharField(
        max_length=255,
        required=True,
        label="Hostname",
    )
    mac = MacField(
        max_length=17,
        required=True,
        label="MAC address",
        validators=[unique_checker_mac]
    )

    def clean(self):
        cleaned_data = super(HostForm, self).clean()
        hostname = cleaned_data.get("hostname")
        group = cleaned_data.get("group")
        host = Host.objects.filter(hostname=hostname, group=group)
        if host:
            raise ValidationError("Host {} already exists in group {}".format(
                hostname, group.pk
            ))


class HostGroupForm(forms.ModelForm):
    name = forms.CharField(max_length=255)

    def __init__(self, *args, **kwargs):
        super(HostGroupForm, self).__init__(*args, **kwargs)
        self.fields['smart_proxy_id'] = forms.ChoiceField(
            choices=[(i.id, i.hostname) for i in SmartProxy.objects.all()],
            initial=self.instance.smart_proxy_id,
            label="Smart Proxy"
        )
        for field in self.fields.values():
                field.widget.attrs['class'] = 'form-control'
        for field in ProfileField.objects.filter(visible_in='group_fields'):
            field_name = field.field_name
            # try to find local Field class definitions
            # and if not found try to get it from forms.*
            field_obj = locals().get(field_name, getattr(forms, field.field))
            field_kwargs = field.params.copy()
            # keep compatibility
            if field.field_name == 'puppet_server':
                field_obj = PuppetMasterField
            elif field.widget:
                field_kwargs['widget'] = getattr(forms, field.widget)
            if self.instance.params and field_name in self.instance.params:
                field_kwargs['initial'] = self.instance.params[field_name]
            self.fields[field.field_name] = field_obj(**field_kwargs)
            self.fields[field.field_name].widget.attrs['class'] = 'form-control'

    def save(self, commit=True):
        instance = super(HostGroupForm, self).save(commit=False)
        if 'smart_proxy_id' in self.cleaned_data:
            instance.smart_proxy_id = int(self.cleaned_data['smart_proxy_id'])
        instance.params = {}

        for field in ProfileField.objects.filter(visible_in='group_fields'):
            instance.params[field.field_name] = self.cleaned_data[field.field_name]
        if commit:
            instance.save()
        return instance

    class Meta:
        model = HostGroup
        fields = ('name', 'source', 'source_params')


class ProfileForm(forms.ModelForm):

    def __init__(self, *args, **kwargs):
        super(ProfileForm, self).__init__(*args, **kwargs)
        self.fields['profile_archive'] = forms.FileField()

    def is_valid(self):
        is_valid = super(ProfileForm, self).is_valid()
        if not is_valid:
            return is_valid
        # Additional checks...
        if 'profile_archive' in self.files:
            from tempfile import NamedTemporaryFile
            from pxetool.exceptions import ImportProfileException
            f = NamedTemporaryFile(delete=False)
            try:
                for chunk in self.files['profile_archive'].chunks():
                    f.write(chunk)
            finally:
                f.close()
            try:
                self.instance.import_archive(f.name, validate=True)
            except ImportProfileException as e:
                self.errors['profile'] = e.message
                is_valid = False
        return is_valid

    class Meta:
        model = Profile
        fields = ('name', )


class SmartProxyForm(forms.ModelForm):

    def __init__(self, *args, **kwargs):
        super(SmartProxyForm, self).__init__(*args, **kwargs)
        for field in self.fields.values():
            field.widget.attrs['class'] = 'form-control'
        ssh_key_path = os.path.join(
            pwd.getpwuid(os.geteuid())[5],
            ".ssh/id_rsa.pub",
        )
        with open(ssh_key_path) as f:
            ssh_key = f.read()
        self.fields['hostname'].help_text = '''
        <script>
        $(function () {{
          $('[data-toggle="popover"]').popover()
        }})</script>
        <p>Put this <span class="label label-default" role="button"
        data-toggle="collapse" href="#collapseExample" aria-expanded="false"
        aria-controls="collapseExample">ssh key</span> under user  on Smart Proxy host</p>
        <pre class="collapse" id="collapseExample">{}</pre>
        '''.format(ssh_key)

    class Meta:
        model = SmartProxy
        fields = ('hostname', 'username', 'pxetool_url', 'tftpboot', 'port')


class HostSourceForm(forms.ModelForm):
    params = forms.CharField(
        label='Auth data',
        widget=forms.Textarea(attrs={'cols': '40', 'rows': '5'})
    )

    def __init__(self, *args, **kwargs):
        super(HostSourceForm, self).__init__(*args, **kwargs)
        for field in self.fields.values():
            if isinstance(field.widget, forms.CheckboxInput):
                field.widget.attrs['class'] = 'form-control'

    class Meta:
        model = HostSource
        fields = ('name', 'url', 'type', 'params', 'enable')


class PuppetMasterForm(forms.ModelForm):
    _text_area = forms.Textarea(attrs={'cols': '40', 'rows': '20'})
    ca_content = forms.CharField(label="CA.crt content", widget=_text_area)
    cert_content = forms.CharField(label="Certificate content", widget=_text_area)
    key_content = forms.CharField(label="Certificate Key content", widget=_text_area)

    def __init__(self, *args, **kwargs):
        super(PuppetMasterForm, self).__init__(*args, **kwargs)
        for field in self.fields.values():
            if field.widget.__class__ != forms.CheckboxInput:
                field.widget.attrs['class'] = 'form-control'

    class Meta:
        model = PuppetMaster
        fields = ['hostname', 'port', 'version', 'ca_content', 'cert_content', 'key_content']
