import os
import pwd

from django.test import TestCase
from django.test.client import Client

from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from pxetool.models import Host, HostGroup, SmartProxy
from pxetool_drivers.models import DriverManager

username = pwd.getpwuid(os.getuid()).pw_name


class LoginTestCase(TestCase):
    def setUp(self):
        self.login_data = {
            'username': 'login_test',
            'password': 'user',
        }
        self.user = User.objects.create_superuser(
            email='user@host.local',
            **self.login_data
        )
        self.c = Client()

    def test_login(self):
        response = self.c.post(reverse('django.contrib.auth.views.login'), data={
            'username': self.login_data['username'],
            'password': self.login_data['password'],
            'next': reverse('index'),
        })
        self.assertRedirects(response, reverse('index'))
        self.assertContains(self.c.get(reverse('index')), 'Welcome, {}.'.format(
            self.login_data['username'],
        ))

    def test_logout(self):
        self.c.login(**self.login_data)
        response = self.c.get(
            reverse('django.contrib.auth.views.logout') +
            "?next=" + reverse('django.contrib.auth.views.login')
        )
        self.assertRedirects(response, reverse('django.contrib.auth.views.login'))

    def test_require_login(self):
        get = [
            reverse('index'),
            reverse('category', kwargs={'active_group': 'default'}),
            reverse('profile_list'),
            reverse('sources_list'),
            reverse('smart_proxy_list'),
        ]
        login_page = reverse('django.contrib.auth.views.login')
        for url in get:
            self.assertRedirects(self.c.get(url), login_page + "?next=" + url)


class NormalScenario(TestCase):
    def setUp(self):
        self.login_data = {
            'username': 'user',
            'password': 'user',
        }
        self.user = User.objects.create_superuser(
            email='user@host.local',
            **self.login_data
        )
        self.c = Client()
        self.c.login(**self.login_data)

    def test_simple(self):
        """Basic scenario:
        - add profile
        - add smart-proxy
        - add group
        - add host
        - begin install
        """
        data = {
            'hostname': 'localhost',
            'username': username,
            'pxetool_url': 'http://localhost:8000',
            'tftpboot': '/tmp',
            'port': '22'
        }
        response = self.c.post(reverse('smart_proxy_create'), data)
        self.assertRedirects(response, reverse('smart_proxy_list'))
        data = {
            'hostname': 'localhost',
            'username': 'root',
            'pxetool_url': 'http://localhost:8000',
            'tftpboot': '/tmp',
            'port': '22'
        }


class NormalUserTestCase(TestCase):
    def setUp(self):
        self.login_data = {
            'username': 'user',
            'password': 'user',
        }
        self.user = User.objects.create_superuser(
            email='user@host.local',
            **self.login_data
        )
        self.c = Client()
        self.smart_proxy = SmartProxy.objects.create(
            hostname='localhost',
            username=username,
        )
        self.hg = HostGroup.objects.create(
            name='default',
            smart_proxy=self.smart_proxy
        )
        self.c.login(**self.login_data)

    def test_get_index(self):
        response = self.c.get(reverse('index'))
        self.assertEqual(response.status_code, 200)

    def test_create_smart_proxy(self):
        response = self.c.get(reverse('smart_proxy_create'))
        self.assertEqual(response.status_code, 200)
        response = self.c.post(reverse('smart_proxy_create'), {
            'hostname': '127.0.0.2',
            'username': username,
            'port': '22',
            'tftpboot': '/tmp',
            'pxetool_url': 'http://127.0.0.2/',
        })
        self.assertRedirects(response, reverse('smart_proxy_list'))

    def test_create_group(self):
        url = reverse('group_create')
        response = self.c.get(url)
        self.assertEqual(response.status_code, 200)
        response = self.c.post(url, {
            'name': 'test_group',
            'smart_proxy_id': self.smart_proxy.id
        })
        self.assertRedirects(
            response,
            reverse('category', kwargs={'active_group': 'test_group'})
        )

    def test_edit_group(self):
        url = reverse('group_edit', kwargs={'pk': self.hg.pk})
        response = self.c.get(url)
        self.assertEqual(response.status_code, 200)
        response = self.c.post(url, {
            'name': 'new test name',
            'smart_proxy_id': self.smart_proxy.id,
        })
        self.assertRedirects(
            response,
            reverse('category', kwargs={'active_group': 'new test name'})
        )

    def validate_host(self, data):
        """check if host has correct data"""
        driver_name = data['driver_name']
        host = Host.objects.filter(
            hostname=data['hostname'],
            mac=data['mac'],
            group_id=data['group'],
        )
        self.assertEqual(host.count(), 1)
        driver = DriverManager.objects.filter(host=host[0], driver_name=driver_name)
        self.assertEqual(driver.count(), 1)
        if data['driver_name'] == 'powerless':
            return
        self.assertEqual(
            driver[0].params,
            {
                k.replace(driver_name + '-', ''): v
                for k, v in data.iteritems() if k.startswith(driver_name)
            }
        )

    def test_create_host_powerless(self):
        url = reverse('host_create')
        data = {
            'hostname': 'test_host_powerless',
            'mac': '00-00-01-01-02-33',
            'group': self.hg.id,
            'driver_name': 'powerless',
        }
        response = self.c.post(url, data)
        self.assertRedirects(
            response,
            reverse('category', kwargs={'active_group': self.hg.name})
        )
        self.validate_host(data)

    def test_create_host_virsh(self):
        url = reverse('host_create')
        data = {
            'hostname': 'test_group',
            'mac': '00-00-01-01-02-02',
            'group': self.hg.id,
            'driver_name': 'virsh',
            'virsh-conn_string': 'qemu:///system',
        }
        response = self.c.post(url, data)
        self.assertRedirects(
            response,
            reverse('category', kwargs={'active_group': self.hg.name})
        )
        self.validate_host(data)

    def test_edit_host(self):
        url = reverse('host_create')
        data = {
            'hostname': 'test_host_powerless',
            'mac': '00-00-01-01-02-33',
            'group': self.hg.id,
            'driver_name': 'powerless',
        }
        self.c.post(url, data)
        host = Host.objects.get(hostname='test_host_powerless')
        data = {
            'hostname': 'test_group',
            'mac': '00-00-01-01-02-02',
            'group': self.hg.id,
            'driver_name': 'virsh',
            'virsh-conn_string': 'qemu:///system',
        }
        url = reverse('host_edit', kwargs={'pk': host.id})
        self.c.post(url, data)
        self.validate_host(data)
