import tempfile
import re
import json

from django.http import HttpResponseRedirect, HttpResponse, HttpResponseNotFound, Http404
from django.db import IntegrityError
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.core.urlresolvers import reverse, reverse_lazy
from django.views.generic import ListView, CreateView, DetailView, UpdateView, View, TemplateView, DeleteView
from django.template.loader import get_template, TemplateDoesNotExist
from django.contrib import messages
from django.forms.models import modelform_factory
from django.shortcuts import render
from django.conf import settings
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt

from braces.views import LoginRequiredMixin

from pxetool.models import Host, Installation, LogFile, HostGroup, Profile, SmartProxy, InstallationFile
from pxetool.models import form_profile_host, HostSource
from pxetool.models import PuppetMaster
from pxetool_drivers.models import DriverManager
from pxetool.exceptions import InstallationException
from .forms import HostForm, HostGroupForm, ProfileForm, SmartProxyForm, \
    HostSourceForm, PuppetMasterForm

from logging import getLogger
logger = getLogger(__name__)


class InstallView(LoginRequiredMixin, View):
    def post(self, request, *args, **kwargs):
        hosts = request.POST.getlist('hosts[]')
        referer = request.META.get('HTTP_REFERER', reverse('index'))

        try:
            profile = Profile.objects.get(pk=int(request.POST.get('profile')))
        except ObjectDoesNotExist:
            return HttpResponseRedirect(referer)
        for pk in hosts:
            try:
                host = Host.objects.get(pk=pk)
                installations = Installation.objects.filter(host=host)
            except ObjectDoesNotExist:
                continue
            if installations.count() > 0:
                install = installations.latest('created')
                if install.status not in ['done', 'error', 'unknown']:
                    installations.latest('created').update_status('aborted')
            try:
                Installation.objects.create(
                    host=host,
                    profile=profile,
                    owner=request.user,
                )
            except InstallationException as e:
                messages.error(request, e.message)
            except IntegrityError as e:
                messages.error(request, "Unknown error: {}".format(e.message))
        return HttpResponseRedirect(referer)


class InstallationBaseView(View):
    """
    This views are used only in scripts during installation and
    scripts fetch this URLs via curl. Becase of that csrf are disabled.
    """
    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return super(InstallationBaseView, self).dispatch(request, *args, **kwargs)


class CfgView(InstallationBaseView):
    """
    Status update view used by debian-installer/anakonda/install.sh
    should be available without auth
    Update: only available without auth to a single IP address, to reduce sniffing chances
    """
    def get(self, request, *args, **kwargs):
        # kept for compatibility with old profiles
        uuid = self.kwargs.get('uuid', self.kwargs.get('hash', ''))

        # For non-authenticated requests, remember IP from first request and demand it afterwards.
        if not request.user.is_authenticated():
            client_ip = request.META['REMOTE_ADDR']
            try:
                installation = Installation.objects.get(uuid=uuid)
            except Installation.DoesNotExist:
                raise HttpResponseNotFound
            if installation.ip is None:
                installation.ip = client_ip
                installation.save(update_fields=['ip'])
            elif installation.ip != client_ip:
                raise PermissionDenied

        try:
            file_obj = InstallationFile.objects.get(
                install__uuid=uuid,
                filename=self.kwargs.get('filename', ''),
            )
        except ObjectDoesNotExist:
            raise HttpResponseNotFound
        if 'preview' not in request.GET:
            if file_obj.status:
                file_obj.install.update_status(file_obj.status)
        return HttpResponse(
            file_obj.content,
            content_type="text/plain"
        )


class StatusUpdateView(InstallationBaseView):
    """
    Status update view used by install.sh and it
    should be available without auth
    """
    def post(self, request, *args, **kwargs):
        error = status = ''
        # kept for compatibility with old profiles
        uuid = self.kwargs.get('uuid', self.kwargs.get('hash', ''))
        try:
            installation = Installation.objects.get(uuid=uuid)
            if 'status' in request.POST:
                status = request.POST.get('status')
                installation.update_status(status)
        except ObjectDoesNotExist:
            status = 'error'
            error = 'not found'
        return HttpResponse(
            json.dumps({'status': status, 'error': error}),
            content_type="application/json"
        )


class AbortInstallation(LoginRequiredMixin, View):
    def post(self, request, *args, **kwargs):
        referer = request.META.get('HTTP_REFERER', reverse('index'))
        hosts = request.POST.getlist('hosts[]')
        hosts = Host.objects.filter(pk__in=hosts)
        for host in hosts:
            if host.installing:
                host.installing.abort()
        return HttpResponseRedirect(referer)


class RecieveLog(InstallationBaseView):
    """
    Fetch log from install.sh script and it
    should be available without auth
    """
    def post(self, request, *args, **kwargs):
        # kept for compatibility with old profiles
        uuid = self.kwargs.get('uuid', self.kwargs.get('hash', ''))
        error = status = ''
        try:
            install = Installation.objects.get(uuid=uuid)
            logfile = request.FILES.get('logfile')
            if logfile:
                log_file = LogFile(installation=install)
                log_file.filename = logfile.name
                log_file.content = logfile.read()
                log_file.save()
                status = 'ok'
            else:
                status = 'error'
                error = 'no log specified'
        except ObjectDoesNotExist:
            status = 'error'
            error = 'not found'
        return HttpResponse(
            json.dumps({'status': status, 'error': error}),
            content_type="application/json"
        )


class ShowLog(LoginRequiredMixin, View):

    def get(self, request, *args, **kwargs):
        # kept for compatibility with old profiles
        uuid = self.kwargs.get('uuid', self.kwargs.get('hash', ''))
        try:
            log_file = LogFile.objects.get(
                installation__uuid=uuid,
                pk=kwargs.get('id'),
            )
        except ObjectDoesNotExist:
            raise HttpResponseNotFound
        return HttpResponse(
            log_file.content,
            content_type="text/plain"
        )


class CreateHostView(LoginRequiredMixin, CreateView):
    form_class = HostForm

    def post(self, request, *args, **kwargs):
        post = request.POST
        form = self.form_class(post)
        if form.is_valid():
            host = Host.create(post)
            return HttpResponseRedirect(reverse('category', args=(host.group.name,)))
        return render(request, 'hosts/index.html', {'form': form})


class HostDetail(LoginRequiredMixin, DetailView):
    model = Host
    template_name = 'hosts/host.html'


class DeleteHostView(LoginRequiredMixin, View):
    def post(self, request, *args, **kwargs):
        hosts = request.POST.getlist('hosts[]')
        Host.objects.filter(pk__in=hosts).delete()
        return HttpResponseRedirect(request.META.get('HTTP_REFERER', reverse('index')))


class UpdateHostView(LoginRequiredMixin, UpdateView):
    model = Host
    form = HostForm
    context_object_name = "host"
    fields = ('hostname', 'mac', 'group')
    template_name = 'hosts/host_update.html'

    def post(self, request, *args, **kwargs):
        post = request.POST.copy()
        if post.get('driver_name') in settings.DRIVER_LIST:
            driver_name = request.POST['driver_name']
        else:
            driver_name = 'powerless'
        host = self.object = self.get_object()
        host.driver.driver_name = driver_name

        # Getting form classes
        host_class = modelform_factory(
            Host, fields=self.fields
        )
        host_form = host_class(data=post, instance=host)
        real_driver_form = DriverManager.get_form(
            driver_name=driver_name,
            data=post
        )

        profile_form = form_profile_host(data=post)

        if host_form.is_valid() and real_driver_form.is_valid() and profile_form.is_valid():
            host.params = profile_form.cleaned_data
            host.save()
            host.driver.params = real_driver_form.cleaned_data
            host.driver.save()
            return super(UpdateHostView, self).post(request, *args, **kwargs)
        else:
            return self.render_to_response(self.get_context_data(form=host_form))

    def get_success_url(self):
        return reverse('category', args=(self.object.group.name,))


class HostGroupMixin(LoginRequiredMixin):
    model = HostGroup


class HostGroupFormMixin(HostGroupMixin):
    form_class = HostGroupForm


class HostGroupView(HostGroupMixin, ListView):
    template_name = 'hosts/index.html'
    context_object_name = 'groups'

    def get_queryset(self):
        q = super(HostGroupView, self).get_queryset()
        return q.select_related('smart_proxy')

    def get_context_data(self, **kwargs):
        context = super(HostGroupView, self).get_context_data(**kwargs)
        # prefetch all data to avoid sql with limit 1 on line
        # >> context['active_group'] = context['groups'][0].name
        context['groups'] = [i for i in self.object_list.all()]
        if 'active_group' in self.kwargs:
            context['active_group'] = self.kwargs['active_group']
        elif len(context['groups']) > 0:
            context['active_group'] = context['groups'][0].name
        return context


class HostGroupDeleteView(HostGroupFormMixin, DeleteView):
    success_url = reverse_lazy('index')


class HostGroupCreateView(HostGroupFormMixin, CreateView):
    template_name = 'hosts/group_create.html'

    def get_success_url(self):
        return reverse('category', args=(self.object.name,))


class HostGroupUpdate(HostGroupFormMixin, UpdateView):
    template_name = 'hosts/host_group_detail.html'

    def get_success_url(self):
        return reverse('category', args=(self.object.name,))

    def get_context_data(self, **kwargs):
        context = super(HostGroupUpdate, self).get_context_data(**kwargs)
        try:
            context['active_group'] = kwargs['active_group']['object'].name
        except KeyError:
            pass
        return context


# Installation Profile views
class ProfileMixin(LoginRequiredMixin):
    model = Profile


class ProfileList(ProfileMixin, ListView):
    pass


class ProfileImport(ProfileMixin, UpdateView):
    form_class = ProfileForm
    success_url = reverse_lazy('index')
    template_name = 'profile_form.html'

    def get_object(self, queryset=None):
        try:
            if 'pk' in self.kwargs:
                return Profile.objects.get(pk=self.kwargs['pk'])
            elif self.request.method == 'POST':
                return Profile.objects.get(name=self.request.POST.get('name', ''))
        except ObjectDoesNotExist:
            return None

    def form_valid(self, form):
        redirect_url = super(ProfileImport, self).form_valid(form)
        if 'profile_archive' in self.request.FILES:
            f = tempfile.NamedTemporaryFile(delete=False)
            for chunk in self.request.FILES['profile_archive'].chunks():
                f.write(chunk)
            f.close()
            self.object.import_archive(f.name)
        return redirect_url


# Docs view
class DocsView(TemplateView):
    """Docs should be available without auth"""
    def get_template_names(self):
        page = self.kwargs.get('page', '')
        if not re.match('^[A-Za-z0-9\-_]+$', page):
            page = 'index'
        template = 'docs/{}.html'.format(page)
        try:
            get_template(template)
            return template
        except TemplateDoesNotExist:
            raise Http404


# Smart Proxy views
class SmartProxyMixin(LoginRequiredMixin):
    model = SmartProxy


class SmartProxyFormMixin(SmartProxyMixin):
    form_class = SmartProxyForm
    success_url = reverse_lazy('smart_proxy_list')


class SmartProxyList(SmartProxyMixin, ListView):
    pass


class SmartProxyCreate(SmartProxyFormMixin, CreateView):
    pass


class SmartProxyUpdate(SmartProxyFormMixin, UpdateView):
    pass


class SmartProxyDelete(SmartProxyFormMixin, DeleteView):
    pass


# Source of hosts views
class HostSourceMixin(LoginRequiredMixin):
    model = HostSource


class HostSourceFormMixin(HostSourceMixin):
    form_class = HostSourceForm
    success_url = reverse_lazy('sources_list')


class HostSourceList(HostSourceMixin, ListView):
    pass


class HostSourceUpdate(HostSourceFormMixin, UpdateView):
    pass


class HostSourceDelete(HostSourceFormMixin, DeleteView):
    pass


class HostSourceCreate(HostSourceFormMixin, CreateView):
    pass


class HostSourceEnable(HostSourceFormMixin, UpdateView):

    def get(self, request, *args, **kwargs):
        super(HostSourceEnable, self).get(request, *args, **kwargs)
        self.object.enable = not self.object.enable
        self.object.save()
        return HttpResponseRedirect(self.get_success_url())


class PuppetMasterMixin(LoginRequiredMixin):
    model = PuppetMaster


class PuppetMasterFormMixin(PuppetMasterMixin):
    form_class = PuppetMasterForm
    success_url = reverse_lazy('puppet_master_list')


class PuppetMasterList(PuppetMasterMixin, ListView):
    pass


class PuppetMasterUpdate(PuppetMasterFormMixin, UpdateView):
    pass


class PuppetMasterDelete(PuppetMasterFormMixin, DeleteView):
    pass


class PuppetMasterCreate(PuppetMasterFormMixin, CreateView):
    pass
